# -*- coding: utf-8 -*-
# author: pengr

from datacommon.util import createInstance
from datacommon.constants import non
import hashlib

# 对相应的类容执行函数


class DataSet(object):

    def __init__(self, data, nrows=None, ncols=None):
        # 初始化数据集必须为二维列表
        self.__nrows = len(data) if not nrows else nrows
        self.__ncols = len(data[0]) if not ncols else ncols
        self.__data = data
        fill_flag = False
        if not ncols:
            for row in data:
                fill_flag = True if len(
                    row) != self.__ncols else False or fill_flag
                if len(row) > self.__ncols:
                    self.__ncols = len(row)
        if fill_flag:
            self.__fill(self.__nrows, self.__ncols)

    def runUDF(self, UDfunc, *cols, **kwargs):
        moduleName = kwargs['module'] if 'module' in kwargs else 'datacommon.udfs'
        UDfunc = createInstance(moduleName, UDfunc)
        cols = self.__check(cols)
        items = []
        for row in self.__data:
            params = []
            for col in cols:
                params.append(row[col])
            items.append(UDfunc.evaluate(*params, **kwargs))
        new_task = DataSet(items)
        return new_task

    def filterNonData(self, col):
        items = []
        for item in self.__data:
            if item[col] != non:
                items.append(item[col])

        new_task = DataSet(items)
        return new_task

    def distinct(self, *cols):
        cols = self.__check(cols)
        tempSet = set()
        repestIndex = set()
        for rowIndex in range(len(self.__data)):
            row = self.__data[rowIndex]
            digest = self.__digest([row[col] for col in cols])
            if digest in tempSet:
                repestIndex.add(rowIndex)
            else:
                tempSet.add(digest)
        new_task = DataSet([self.__data[i] for i in range(len(self.__data)) if i not in repestIndex])
        return new_task

    def count(self, *cols):
        cols = self.__check(cols)
        res = [0 for col in cols]
        for item in self.__data:
            for idx in range(len(cols)):
                res[idx] = res[idx]+1 if item[cols[idx]
                                              ] != non else res[idx]

        new_task = DataSet([res])
        return new_task

    def min(self, *cols):
        cols = self.__check(cols)
        res = [non for col in cols]
        for item in self.__data:
            for idx in range(len(cols)):
                res[idx] = item[cols[idx]] if res[idx] == non or item[cols[idx]
                                                                      ] < res[idx] else res[idx]

        new_task = DataSet([res])
        return new_task

    def max(self, *cols):
        cols = self.__check(cols)
        res = [non for col in cols]
        for item in self.__data:
            for idx in range(len(cols)):
                res[idx] = item[cols[idx]] if res[idx] == non or item[cols[idx]
                                                                      ] > res[idx] else res[idx]

        new_task = DataSet([res])
        return new_task

    @property
    def data(self):
        return self.__data

    @property
    def nrows(self):
        return self.__nrows

    @property
    def ncols(self):
        return self.__ncols

    def __str__(self):
        res = '<DataSet object \'s data>: {}行, {}列\n'.format(
            self.__nrows, self.__ncols)
        for item in self.__data:
            res += ', '.join([str(i) for i in item]) + '\n'
        return res

    def __check(self, param_cols):
        if param_cols and len(param_cols) > 0:
            return param_cols
        return [i for i in range(self.__ncols)]

    def __digest(self, obj):
        md5 = hashlib.md5()
        md5.update(obj.__str__().encode('utf-8'))
        return md5.hexdigest()[8:-8]

    def __fill(self, nrows, ncols):
        while len(self.__data) < nrows:
            self.__data.append([])
        for row_item in self.__data:
            while len(row_item) < ncols:
                row_item.append(non)


if __name__ == '__main__':
    numberData = [
        [1, 2, 3, 4, 5],
        [2, 4, 6, 8, 10],
        [3, 6, 9, 12, 4]
    ]

    stringData = [
        ['asda', 'asda', 'dwqwe', 'qweq'],
        ['weqcas', 'defsw', 'deqwa', 'ewqwq'],
        ['weqcas', 'defsw', 'deqwa', 'ewqwq']
    ]

    nonTestData = [
        ['asda'],
        ['weqcas', 'defsw', 'deqwa', 'ewqwq'],
        [non, 'none', 'null']
    ]

    task3 = DataSet(stringData)
    print(task3.distinct())

    # task1 = DataSet(numberData)
    # task1.min(1, 3)
    # print(task1)

    # task2 = DataSet(nonTestData)
    # task2.count(0, 1, 2, 3)
    # print(task2)
    # print(task2.count(0, 1, 2, 3).runUDF('PerfectionUDF', nrows=task2.nrows))



# -*- coding: utf-8 -*-
# author: pengr


class FunctionManager:
    def __init__(self):
        print("初始化")
        self.functions = {}

    def executeFunc(self,func=None):
        print(self.functions)
 
    def register(self, func):
        self.functions[func.__name__] = func
 
 
fm = FunctionManager()
 
 
@fm.register
def t1():
    print("t1")
 
 
@fm.register
def t2():
    print("t2")
 
 
@fm.register
def t3():
    print("t3")
 
fm.execute_all()
fm.executeFunc()


# -*- coding: utf-8 -*-
# author: pengr


class NonData():
    # 空数据对象
    __fillvalue = '<NonData object>'

    @property
    def fillvalue(self):
        return self.__fillvalue

    @fillvalue.setter
    def fillValue(self, value):
        self.__fillvalue = value

    def __str__(self):
        return self.__fillvalue


class LastColumn():

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)


non = NonData()



# -*- coding: utf-8 -*-
# author: pengr


import chardet
import codecs

# Null bytes; no need to recreate these on each call to guess_json_utf
_null = '\x00'.encode('ascii')  # encoding to ASCII for Python 3
_null2 = _null * 2
_null3 = _null * 3


def guessTextEncode(text):
    """
    :rtype: str
    """
    res = chardet.detect(text)["encoding"]
    if res:
        return res
    sample = text[:4]
    if sample in (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE):
        return 'utf-32'     # BOM included
    if sample[:3] == codecs.BOM_UTF8:
        return 'utf-8-sig'  # BOM included, MS style (discouraged)
    if sample[:2] in (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE):
        return 'utf-16'     # BOM included
    nullcount = sample.count(_null)
    if nullcount == 0:
        return 'utf-8'
    if nullcount == 2:
        if sample[::2] == _null2:   # 1st and 3rd are null
            return 'utf-16-be'
        if sample[1::2] == _null2:  # 2nd and 4th are null
            return 'utf-16-le'
        # Did not detect 2 valid UTF-16 ascii-range characters
    if nullcount == 3:
        if sample[:3] == _null3:
            return 'utf-32-be'
        if sample[1:] == _null3:
            return 'utf-32-le'
        # Did not detect a valid UTF-32 ascii-range character
    return None



import os

# 根据数据集建立目录
def saveContent(path, content):
    if os.path.exists(path):
        return
    dir = os.path.dirname(path)
    if not os.path.exists(dir):
        os.makedirs(dir)
    with open(path, 'wb') as f:
        f.write(content)

# 罗列文件
def listFile(dir='source'):
    # 读取source文件夹下所有文件
    # @param: str
    file_list = []
    for root, _, files in os.walk(dir):
        for file in files:
            file_list.append(os.path.join(root, file))
    return file_list


# 判断空数据
def isBlank(item, NonKeys=['NONE', 'NULL', 'NON']):
    if item is None or len(str(item)) == 0 or str(item).strip().upper() in NonKeys:
        return True
    for ch in str(item):
        if ch != ' ':
            return False
    return True

# 统计非法日期格式格式数据条目/所占比


import time

# 判断日期格式


def isDateTrueFormat(date, format='%Y-%m-%d'):
    try:
        time.strptime(date, "%Y-%m-%d")
        return True
    except:
        return False


# 通过类名获取对象
def createInstance(module_name, class_name, *args, **kwargs):
    module_meta = __import__(module_name, globals(), locals(), [class_name])
    class_meta = getattr(module_meta, class_name)
    obj = class_meta(*args, **kwargs)
    return obj


# -*- coding: utf-8 -*-
# author: pengr

from . import util
from .constants import non
import re


class BaseUDF(object):
    # 用户自定义函数
    # 结果数据集格式为二维列表
    def evaluate(self, *args, **kwargs):
        raise NotImplementedError()


class TestUDF(BaseUDF):
    # UDF测试用例
    def evaluate(self, *args, **kwargs):
        print('test udf and params is arg: '+args)


class IsNonDataUDF(BaseUDF):
    # 统计数据集中空数据条目/所占比例
    def evaluate(self, *items):
        res = []
        for item in items:
            if util.isBlank(item):
                res.append(non)
            else:
                res.append(item)
        return res


class PerfectionUDF(BaseUDF):
    # 统计数据集中空数据条目/所占比例
    def evaluate(self,  *items, nrows):
        res = [count/nrows for count in items]
        res.append(sum(res)/len(res))
        return res


class GetDateColumns(BaseUDF):
    # 判断列是否为日期列

    __dateKey = ['update', 'create', 'time', '时间', '日期', 'date']

    def evaluate(self, *cols):
        res = set()
        for col in cols:
            if not col:
                continue
            for key in self.__dateKey:
                if key in str(col):
                    res.add(cols.index(col))
            if isinstance(col, int) and 1971 < col < 2050:
                res.add(cols.index(col))
            if re.search('^(^(\d{4}|\d{2})(\-|\/|\.)\d{1,2}\3\d{0,2})|(^\d{4}年\d{1,2}月\d{0,2}日*)', str(col)):
                res.add(cols.index(col))
        return [list(res)]


class IsTruthDateFormatUDF(BaseUDF):

    # 统计日期格式

    def evaluate(self, *cols, format):
        res = []
        for col in cols:
            if util.isDateTrueFormat(col, format=format):
                res.append(col)
            else:
                res.append(non)
        return res


class DropInvalidColumnsUDF(BaseUDF):
    # 取出无效数据列
    def evaluate(self, *items):
        last = len(items)-1
        while util.isBlank(items[last]) and last >= 0:
            last -= 1
        res = list(items[:last])
        return res


# if __name__ == '__main__':
#     print(100)
#     for i in range(10,0, -1):
#         print(i)



# -*- coding: utf-8 -*-
# author: pengr
#               |_|               |_|


'''
通用数据处理工具

:copyright: (c) 2019 by pengr.
:license: GNU GENERAL PUBLIC LICENSE, see LICENSE for more details.
'''

__title__ = '通用数据处理工具'
__description__ = '通用数据处理工具'
__author__ = 'pengr'
__author_email__ = 'pengrui55555@163.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2019 Pengr'

from . import util
from .constants import non
from .reader import getCsvData, getXlsData
from .dataSet import DataSet



# -*- coding: utf-8 -*-
# author: pengr

from .util import guessTextEncode


def getFileData(filename,):
    suffix = filename.split('.')[-1].upper()
    if suffix == 'XLS' or suffix == 'XLSX':
        return getXlsData(filename)
    elif suffix == 'CSV':
        with open(filename, 'rb') as f:
            encoding = guessTextEncode(f.read())
        if encoding:
            return getCsvData(filename, encoding=encoding)
        return getCsvData(filename)
    elif suffix == 'JSON':
        with open(filename, 'rb') as f:
            encoding = guessTextEncode(f.read())
        if encoding:
            return getCsvData(filename, encoding=encoding)
        return getJsonData(filename, encoding=encoding)


import csv


def getCsvData(filename, encoding='utf-8', delimiter=','):
    # csv文件中许多无效列
    with open(filename, encoding=encoding) as f:
        items = [row for row in csv.reader(f, delimiter=delimiter)]
    return items


import xlrd


def getXlsData(filename):
    items = []
    data = xlrd.open_workbook(filename)
    table = data.sheets()[0]
    nrows = table.nrows
    for row in range(nrows):
        items.append(table.row_values(row))
    return items


import json


def getJsonData(filename, encoding='utf-8'):
    fp = open(filename, 'r', encoding=encoding)
    return json.load(fp)


def getTxtData(filename, encoding='utf-8'):
    with open(filename, 'r', encoding=encoding) as f:
        content = f.read()
    return content


from lxml import etree


def getXmlData(filename, encoding='utf-8'):
    data = []
    docTree = etree.parse(filename, etree.XMLParser())
    root = docTree.getroot()
    items = root.getchildren()
    for item in items:
        data.append([i.text for i in item.getchildren()])
    return data


if __name__ == '__main__':
    filename = '..\guizhou\source\贵州省PPP推进会推介项目表\贵州省PPP推进会推介项目表.csv'
    xlsfilename = '..\guizhou\source\部门决算及“三公”经费决算信息公开表\贵州省投资促进局2016年度部门预算及“三公”经费预算.xlsx'
    # data = getCsvData(filename)
    data = getFileData(xlsfilename)

    print(data)

    from .dataSet import DataSet

    ds1 = DataSet(data)
    cols = [i for i in range(9)]
    date_cols = ds1.runUDF('GetDateColumns').distinct()
    print(date_cols)
    x = ds1.runUDF('IsNonDataUDF', *cols).count(*cols)
    print(x)
